// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Xml;
using PiDuce.Common;
using PiDuce.Machine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Diagnostics;

namespace PiDuce.Web
{
  public class Webclient
  {	
    public const string USER_AGENT = "PiDuce";

    public delegate String DownloadDelegate (string url);

    public static String Download(String url)
    {
      WebClient wc = new WebClient();
      wc.Headers.Add("User-Agent", USER_AGENT);
      return System.Text.Encoding.UTF8.GetString(wc.DownloadData(url));
    }

    public delegate XmlNodeList UploadDelegate (String url, String data, String soapAction);

    public XmlNodeList Upload(String url, String data, String soapAction)
    {
      System.Net.WebClient wc = new WebClient();
      byte[] request = new byte[data.Length];
      request = new System.Text.UTF8Encoding().GetBytes(data);
      wc.Headers.Add("SOAPAction", soapAction);
      wc.Headers.Add("Content-Type", "text/xml; charset=utf-8");
      wc.Headers.Add("User-Agent", USER_AGENT);
      try
      {
        byte[] response = wc.UploadData(url, "POST", request);
        string xml = System.Text.Encoding.UTF8.GetString(response, 0, response.Length);
        return SoapHandler.RemoveSoap(xml);
      }
      catch (WebException e)
      {
        Machine.MachineOutput.Print(e.Message + " " + e.Response.GetResponseStream().ToString());
        throw e;
      }
    }
  }

  public class WebServer
  {
    public const string LINEARFWD = "http://www.cs.unibo.it/PiDuce/LinearForwarder";
    public const string LOADCODE = "http://www.cs.unibo.it/PiDuce/LoadCode";
    public const string ADDCHANNEL = "http://www.cs.unibo.it/PiDuce/AddChannel";

    private static IManageOutputRequest channelManager;
    private static IScheduler scheduler;
    private static string localMachine;
    private static int port;
    private static string machineAbstractWsdl;
    private static bool keepAlive;
    private static ManualResetEvent allDone;

    public static int Port { get { return port;} }
    public static string LocalMachine
    { get { return "http://" + localMachine + ":" + port.ToString() + "/"; } }

    static WebServer()
    {
      WebServer.port = 1811;
      WebServer.keepAlive = true;
      WebServer.allDone = new ManualResetEvent(false);
      WebServer.machineAbstractWsdl = "machineAbstract.wsdl";
    }

    public static void Init(IManageOutputRequest channelManager, IScheduler scheduler, string host, int port)
    {
      WebServer.channelManager = channelManager;
      WebServer.scheduler = scheduler;
      WebServer.port = port;
      try
      {
          IPHostEntry ipHostEntry = Dns.GetHostEntry(host);
          if (ipHostEntry.AddressList.Length == 0)
              WebServer.localMachine = IPAddress.Loopback.ToString();
          else
          {
              foreach (IPAddress ip in ipHostEntry.AddressList)
              {
                  if (ip.AddressFamily == AddressFamily.InterNetwork)
                  {
                      WebServer.localMachine = ip.ToString();
                      break;
                  }
              }
          }
      }
      catch (SocketException)
      {
          WebServer.localMachine = IPAddress.Loopback.ToString();
      }
      catch (ArgumentException)
      {
          WebServer.localMachine = IPAddress.Loopback.ToString();
      }
    }

    public static void Stop()
    {
      keepAlive = false;
      allDone.Set();
    }
    
    public static void StartListening() 
    {
        try
        {
            HttpListener listener = new HttpListener();
            listener.Prefixes.Add(WebServer.LocalMachine);
            try
            {
                listener.Start();
                Console.WriteLine("The machine is waiting on " + WebServer.LocalMachine);
                while (keepAlive)
                {
                    allDone.Reset();
                    listener.BeginGetContext(new AsyncCallback(AcceptCallback), listener);
                    allDone.WaitOne();
                }
                listener.Stop();
            }
            catch (HttpListenerException e) 
            {
                Console.WriteLine("The PiDuce Web server cannot be started. Only local interactions are possible. {0}", e.ToString());                
            }
        }
        catch (PlatformNotSupportedException)
        {
            Console.WriteLine("The PiDuce Web server cannot be started. Only local interactions are possible. (HttpListener class is not available for this platform)");                
        }
    }

    private static void AcceptCallback(IAsyncResult ar) 
    {
      allDone.Set();
      HttpListener listener = (HttpListener)ar.AsyncState;
      HttpListenerContext context = listener.EndGetContext(ar);
      try
      {
        ProcessRequest(context);
      } 
      catch (Exception e) {
        context.Response.StatusCode = (int)HttpStatusCode.InternalServerError;
        context.Response.StatusDescription = e.Message;
        context.Response.Close();
      }
    }
    
    private static string BodyAsString(HttpListenerRequest request)
    {
      StringBuilder builder = new StringBuilder();
      Stream stream = request.InputStream;
      if (stream.CanSeek)
      {
        for (int read = 0; read < request.ContentLength64; )
        {
          int length = (int)stream.Length;
          byte[] received = new byte[length];
          stream.Read(received, 0, length);
          read += length;
          builder.Append(received);
        }
      }
      else {
        for (int read = 0; read < request.ContentLength64; read++)
        {
          byte[] b = new byte[1];
          stream.Read(b, 0, 1);
          builder.Append(System.Text.Encoding.UTF8.GetChars(b));
        }
      }
      return builder.ToString();
    }

    private static void ProcessLinearForwarder(HttpListenerRequest request, HttpListenerResponse response)
    {
      try
      {
        string body = BodyAsString(request);
        XmlNode req = XmlUtil.RemoveSoap(body)[0];
        //MachineOutput.Print("Have received a linearforwarder's request");
        string source = req.ChildNodes[0].InnerText;
        string destination = req.ChildNodes[1].InnerText;
        if (channelManager.LinearForwarder(source, destination))
        {
          response.StatusCode = (int)HttpStatusCode.OK;
          response.ContentLength64 = 0;
        }
        else
        {
          response.StatusCode = (int)HttpStatusCode.NotFound;
          response.ContentLength64 = 0;
        }
      } 
      catch(Exception e)
      {
        response.StatusCode = (int)HttpStatusCode.BadRequest;
        response.ContentLength64 = 0;
        response.StatusDescription = e.Message;
      }
      response.Close();
    }

    private static void ProcessLoadCode(HttpListenerRequest request, HttpListenerResponse response)
    {
      try
      {
        MachineOutput.Print("Have received remote code i start loading it");
        scheduler.AddProgram(XmlUtil.RemoveSoap(BodyAsString(request))[0].OuterXml);
        response.StatusCode = (int)HttpStatusCode.OK;
        response.ContentLength64 = 0;
      }
      catch (Exception e) 
      {
        response.StatusCode = (int)HttpStatusCode.BadRequest;
        response.ContentLength64 = 0;
        response.StatusDescription = "The bytecode may be not correct. " + e.Message;
      }
      response.Close();
    }

    private static void ProcessAddService(HttpListenerRequest request, HttpListenerResponse response)
    {
      MachineOutput.Print("Have received request to add channel  ");
      try
      {
        XmlNode addChannelNode = XmlUtil.RemoveSoap(BodyAsString(request))[0];
        string location;
        if (addChannelNode.ChildNodes[0].LocalName == "location")
        {
          string suggestedLocation = addChannelNode.ChildNodes[0].InnerText;
          string wsdl = addChannelNode.ChildNodes[1].OuterXml;
          ServiceSchema serviceSchema = AbstractWsdlReader.CompileAbstractWsdl(wsdl);
          location = channelManager.AddService(suggestedLocation, serviceSchema);
        }
        else 
        {
          string wsdl = addChannelNode.ChildNodes[0].InnerXml;
          ServiceSchema serviceSchema = AbstractWsdlReader.CompileAbstractWsdl(wsdl);
          location = channelManager.AddService(serviceSchema);
        }
        string message = XmlUtil.AddSoap("<Service xmlns=\"http://www.cs.unibo.it/PiDuce/Default\">" + location + "?wsdl</Service>");
        response.StatusCode = (int)HttpStatusCode.OK;
        response.ContentLength64 = message.Length;
        response.ContentEncoding = System.Text.Encoding.UTF8;
        byte[] bytes = System.Text.Encoding.UTF8.GetBytes(message);
        response.OutputStream.Write(bytes, 0, bytes.Length);
      }
      catch (Exception e)
      {
        response.StatusCode = (int)HttpStatusCode.BadRequest;
        response.ContentLength64 = 0;
        response.StatusDescription = "The Wsdl may be not correct. " + e.Message;
      }
      response.Close();
    }

    private static void ProcessServiceInvocation(HttpListenerRequest request, HttpListenerResponse response) 
    {
      try
      {
        //Post on a local channel
        string location = request.Headers["SOAPAction"];
        if (location.StartsWith("\""))
          location = location.Remove(0, 1);
        if (location.EndsWith("\""))
          location = location.Remove(location.Length - 1, 1);
        ArrayList responsecontainer = new ArrayList();

        LiteralValueUtil valueUtil = new LiteralValueUtil();
        ISchema opSchema = channelManager.GetSchemaOfChannel(location);
        ISchema valueSchema;
        if (opSchema is FunctionSchema)
          valueSchema = ((FunctionSchema)opSchema).Input;
        else
          valueSchema = ((ChannelSchema)opSchema).Content;
        XmlNodeList nodes = SoapHandler.RemoveSoap(BodyAsString(request));
        if (request.UserAgent != Webclient.USER_AGENT && !(valueSchema is LabelledSchema) && !(valueSchema is VoidSchema))
          nodes = nodes.Item(0).ChildNodes;
        IValue val = valueUtil.XmlToIValue(nodes, valueSchema);
        if (channelManager.ForeignRequest(location, val, responsecontainer))
        {
          lock (responsecontainer)
          {
            //to prevent deadlock
            if (responsecontainer.Count == 0)
              System.Threading.Monitor.Wait(responsecontainer);
            IValue v = (IValue)responsecontainer[0];
            //add the wrapper
            ISchema outSchema = ((FunctionSchema)opSchema).Output;
            if (request.UserAgent != Webclient.USER_AGENT && !(outSchema is LabelledSchema) && !(outSchema is VoidSchema))
              v = new LabelValue(channelManager.GetOperationName(location) + "Out", v);
            string message = ValueUtil.MakeDocumentLiteralEncoding(
              channelManager.GetSchemaOfChannel(location), 
              WebNamespace.PiDuceDefaultNamespace)(v);
            response.StatusCode = (int)HttpStatusCode.OK;
            response.ContentType = "text/xml";
            response.ContentEncoding = System.Text.Encoding.UTF8;
            response.ContentLength64 = message.Length;
            byte[] bytes = System.Text.Encoding.UTF8.GetBytes(message);
            response.OutputStream.Write(bytes, 0, bytes.Length);
          }
        }
        else if (channelManager.LocalOutput(location, val))
        {
          response.StatusCode = (int)HttpStatusCode.OK;
          response.ContentLength64 = 0;
        }
        else
        {
          response.StatusCode = (int)HttpStatusCode.NotFound;
          response.ContentLength64 = 0;
        }
      }
      catch (Exception e) 
      {
        response.StatusCode = (int)HttpStatusCode.BadRequest;
        response.ContentLength64 = 0;
        response.StatusDescription = e.Message;
      } 
      response.Close();
    }

    private static void ProcessGetServiceInterface(HttpListenerRequest request, HttpListenerResponse response)
    {
      try
      {
        if (request.RawUrl == "?wsdl" || request.RawUrl == "/?wsdl")
        {
          string service = "  <wsdl:service name=\"PiDuceMachineService\">";
          service += System.Environment.NewLine + "    <wsdl:port name=\"PiDuceMachine\" binding=\"tns:PiDuceMachine\">";
          service += System.Environment.NewLine + "      <soap:address location=\"" + WebServer.LocalMachine + "\" />";
          service += System.Environment.NewLine + "    </wsdl:port>";
          service += System.Environment.NewLine + "  </wsdl:service>";
          service += System.Environment.NewLine + "</wsdl:definitions>";
          //select the abstract wsdl file in the directory of this executable
          string machineExecDirectory = Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location);
          FileStream stream = new FileStream(machineExecDirectory + Path.AltDirectorySeparatorChar + machineAbstractWsdl, FileMode.Open, FileAccess.Read);
          byte[] wsdl = new byte[stream.Length];
          stream.Read(wsdl, 0, (int)stream.Length);
          stream.Close();
          string message = System.Text.Encoding.UTF8.GetString(wsdl, 0, wsdl.Length);
          message = message.Replace("</wsdl:definitions>", service);
          byte[] bytes = System.Text.Encoding.UTF8.GetBytes(message);
          response.StatusCode = (int)HttpStatusCode.OK;
          response.ContentLength64 = bytes.Length;
          response.ContentEncoding = System.Text.Encoding.UTF8;
          response.ContentType = "text/xml";
          response.OutputStream.Write(bytes, 0, bytes.Length);
        }
        else
        {
          string location = request.Url.AbsoluteUri.Substring(0, request.Url.AbsoluteUri.Length - 5);
          if (!request.Url.AbsoluteUri.EndsWith("?wsdl") || !channelManager.ContainsService(location))
          {
            response.StatusCode = (int)HttpStatusCode.NotFound;
            response.ContentLength64 = 0;
          }
          else
          {
            string portTypeName = String.Empty;
            if (request.RawUrl.Contains("?wsdl"))
              portTypeName = request.RawUrl.Replace("?wsdl", "");
            else if (request.RawUrl.Contains("/?wsdl"))
              portTypeName = request.RawUrl.Replace("/?wsdl", "");
            if (portTypeName.StartsWith("/")) 
              portTypeName = portTypeName.Remove(0,1);
            string targetNamespace = WebNamespace.PiDuceDefaultNamespace;
            ServiceSchema serviceSchema = channelManager.GetSchemaOfService(location);
            String message = WsdlGenerator.Generate(serviceSchema, portTypeName, targetNamespace, WebServer.LocalMachine);
            response.StatusCode = (int)HttpStatusCode.OK;
            response.ContentLength64 = message.Length;
            response.ContentEncoding = System.Text.Encoding.UTF8;
            response.ContentType = "text/xml";
            byte[] bytes = System.Text.Encoding.UTF8.GetBytes(message);
            response.OutputStream.Write(bytes, 0, bytes.Length);
          }
        }
      }
      catch (Exception e)
      {
        response.StatusCode = (int)HttpStatusCode.InternalServerError;
        response.ContentLength64 = 0;
        response.StatusDescription = e.Message;
      }
      response.Close();
    }

    private static void ProcessRequest(HttpListenerContext context)
    {
      HttpListenerRequest request = context.Request;
      HttpListenerResponse response = context.Response;
      if (request.HttpMethod == "POST")
      {
        string soapAction = request.Headers["SOAPAction"];
        soapAction = soapAction.Replace("\"", "");
        if (soapAction == LINEARFWD)
          ProcessLinearForwarder(request, response);
        else if (soapAction == LOADCODE)
          ProcessLoadCode(request, response);
        else if (soapAction == ADDCHANNEL)
          ProcessAddService(request, response);
        else ProcessServiceInvocation(request, response);
      }
      else if (request.HttpMethod == "GET")
        ProcessGetServiceInterface(request, response);
      else
      {
        response.StatusCode = (int)HttpStatusCode.NotImplemented;
        response.ContentLength64 = 0;
        response.Close();
      }
    }	
  }
}
