// This file is part of the PiDuce project (http://www.cs.unibo.it/Piduce/)
// developed at the Department of Computer Science of Bologna.
// Authors: Samuele Carpineti, Cosimo Laneve, Luca Padovani
//
// See Copyright for the status of this software.

using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using System.Xml;
using PiDuce.Common;
using PiDuce.Web;

namespace PiDuce.Wsdl2PiDuce
{
    /// <summary>
    /// This utility allows to read remote WSDL files producing a PiDuce prototype file.
    /// Such a prototype is just a sequence of "import" directives.
    /// </summary>
    public class Wsdl2PiDuce
    {
        private static string usage =
    @"Wsdl2PiDuce - Usage: Wsdl2PiDuce [WSDLFILE]s -o [OUTPUTFILE]
  WSDLFILE are generic URI http://..., https://..., file://..., ftp://..., ecc.";

        /// <summary>
        /// Import the referenced WSDL printing the PiDuce skeleton of the schemas to the string builder
        /// </summary>
        /// <param name="wsdl">the wsdl file</param>
        /// <param name="s">the string builder where the PiDuce code is written</param>
        /// <remarks>the StringBuilder <code>s</code> is modified</remarks>
        private static XsdImportOutput ImportWsdl(string wsdl, StringBuilder s)
        {
            WebClient wc = new WebClient();
            XsdToSchema xsd = null;
            ServiceSchema schema = null;
            string name = null;
            if (File.Exists(wsdl))
            {
                schema = AbstractWsdlReader.CompileAbstractWsdl(File.OpenRead(wsdl), out xsd);
                char[] separators = { '.', '?' };
                name = Path.GetFileName(wsdl).Split(separators)[0];
            }
            else
            {
                Uri u = new Uri(wsdl);
                if (u.IsFile)
                    schema = AbstractWsdlReader.CompileAbstractWsdl(File.OpenRead(u.AbsolutePath), out xsd);
                else
                {
                    wc.DownloadData(wsdl);
                    schema = AbstractWsdlReader.CompileAbstractWsdl(wc.OpenRead(wsdl), out xsd);
                }
                char[] separators = { '.', '?' };
                name = Path.GetFileName(u.LocalPath).Split(separators)[0];
                if (name == null || name.Trim().Length == 0)
                    name = wsdl;
            }
            if (xsd.SchemaDefs.Count + xsd.GroupDefs.Count + xsd.ElementDefs.Count > 0)
                s.Append("schema ");
            //top level schema definitions
            //if (xsd.SchemaDefs.Count > 0)
            //    s.AppendLine("//Complex Types Definitions");
            foreach (KeyValuePair<XmlQualifiedName, ConstantSchema> kv in xsd.SchemaDefs)
            {
                ToStringSchemaVisitor visitor = new ToStringSchemaVisitor();
                kv.Value.Entry.Schema.Accept(visitor);
                s.AppendLine(kv.Key.Name + "=" + visitor + " ");
                s.Append("and ");
            }
            if (xsd.SchemaDefs.Count > 0 && xsd.GroupDefs.Count == 0)
                s.Replace("and ", "", s.Length - 4, 4);
            //if (xsd.GroupDefs.Count > 0)
            //    s.AppendLine("//Groups Definitions");
            foreach (KeyValuePair<XmlQualifiedName, ConstantSchema> kv in xsd.GroupDefs)
            {
                ToStringSchemaVisitor visitor = new ToStringSchemaVisitor();
                kv.Value.Entry.Schema.Accept(visitor);
                s.AppendLine(kv.Key.Name + "=" + visitor + " ");
                s.Append("and ");
            }
            if (xsd.GroupDefs.Count > 0 && xsd.ElementDefs.Count == 0)
                s.Replace("and ", "", s.Length - 4, 4);
            //if (xsd.ElementDefs.Count > 0)
            //    s.AppendLine("//Elements Definitions");
            foreach (KeyValuePair<XmlQualifiedName, ConstantSchema> kv in xsd.ElementDefs)
            {
                ToStringSchemaVisitor visitor = new ToStringSchemaVisitor();
                kv.Value.Entry.Schema.Accept(visitor);
                s.AppendLine(kv.Key.Name + "=" + visitor + " ");
                s.Append("and ");
            }
            if (xsd.ElementDefs.Count > 0)
                s.Replace("and ", "", s.Length - 4, 4);
            if (xsd.SchemaDefs.Count + xsd.GroupDefs.Count > 0)
                s.Append("in ");
            if (schema.Operations.Count > 0)
            {
                s.Append("import ");
                s.Append(name + " ");
                s.Append(schema);
            }
            s.AppendLine(System.Environment.NewLine + " location=\"" + wsdl + "\"");
            s.AppendLine("in");
            return xsd.Output;
        }
        private static void WriteFile(string fileName, string s, FileMode mode) 
        {
            FileStream f;
            if (!File.Exists(fileName))
                f = File.Create(fileName);
            else f = File.Open(fileName, mode);
            StreamWriter sw = new StreamWriter(f);
            sw.Write(s);
            sw.Close();
        }
        /// <summary>
        /// Allows to import a list of WSDL files producing the correspondent PiDuce prototype.
        /// The resulting code is written to a file.
        /// </summary>
        /// <param name="args">Usage: Wsdl2PiDuce [WSDLFILE]s -o [OUTPUTFILE]</param>
        public static void Main(string[] args)
        {
            if (args.Length < 2)
            {
                Console.WriteLine(Wsdl2PiDuce.usage);
                return;
            }
            else if (args[args.Length - 2] != "-o")
            {
                Console.WriteLine(Wsdl2PiDuce.usage);
                return;
            }
            else
            {
                string outFileName = args[args.Length - 1];
                if (File.Exists(outFileName))
                {
                    string answer = "n";
                    Console.WriteLine(outFileName + " already exists. Do you want to overwrite it? (y/n) - Default is [N]");
                    answer = Console.In.ReadLine().Trim();
                    if ("n".Equals(answer, StringComparison.OrdinalIgnoreCase))
                        return;
                }
                StringBuilder s = new StringBuilder();
                XsdImportOutput output = null;
                s.AppendLine("//Automatically generated code");
                string logFileName = outFileName + ".log";
                if (args.Length - 2 > 0)
                    File.Delete(logFileName);
                for (int i = 0; i < args.Length - 2; i++)
                {
                    string wsdl = args[i];
                    try
                    {
                        output = ImportWsdl(wsdl, s);
                        WriteFile(logFileName, output.ToString(), FileMode.Append);
                        Console.WriteLine("{0} imported. Please see the log \"{1}\" for errors and warnings about the imported schemas.", wsdl, logFileName);
                    }
                    catch (WebException)
                    {
                        Console.WriteLine(wsdl + " - cannot be downloaded");
                    }
                    catch (XmlException)
                    {
                        Console.WriteLine(wsdl + " - errors reading the wsdl description");
                    }
                    catch (Exception e)
                    {
                        Console.WriteLine(wsdl + " - cannot be read " + e);
                    }
                }
                s.AppendLine(@" nil //replace - nil - with your code!");
                WriteFile(outFileName, s.ToString(), FileMode.Truncate);
            }
        }
    }
}